from typing import Optional

from .. import ser
from ..errors import LuaException
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('turtle.')


def inspect_result(rp):
    success = rp.take_bool()
    if not success:
        msg = rp.take_string()
        if msg == 'No block to inspect':
            return None
        raise LuaException(msg)
    else:
        return rp.take_dict()


__all__ = (
    'craft',
    'forward',
    'back',
    'up',
    'down',
    'turnLeft',
    'turnRight',
    'select',
    'getSelectedSlot',
    'getItemCount',
    'getItemSpace',
    'getItemDetail',
    'equipLeft',
    'equipRight',
    'attack',
    'attackUp',
    'attackDown',
    'dig',
    'digUp',
    'digDown',
    'place',
    'placeUp',
    'placeDown',
    'detect',
    'detectUp',
    'detectDown',
    'inspect',
    'inspectUp',
    'inspectDown',
    'compare',
    'compareUp',
    'compareDown',
    'compareTo',
    'drop',
    'dropUp',
    'dropDown',
    'suck',
    'suckUp',
    'suckDown',
    'refuel',
    'getFuelLevel',
    'getFuelLimit',
    'transferTo',
)


def craft(quantity: int = 64):
    return method('craft', quantity).check_bool_error()


def forward():
    return method('forward').check_bool_error()


def back():
    return method('back').check_bool_error()


def up():
    return method('up').check_bool_error()


def down():
    return method('down').check_bool_error()


def turnLeft():
    return method('turnLeft').check_bool_error()


def turnRight():
    return method('turnRight').check_bool_error()


def select(slotNum: int):
    return method('select', slotNum).check_bool_error()


def getSelectedSlot() -> int:
    return method('getSelectedSlot').take_int()


def getItemCount(slotNum: int = None) -> int:
    return method('getItemCount', slotNum).take_int()


def getItemSpace(slotNum: int = None) -> int:
    return method('getItemSpace', slotNum).take_int()


def getItemDetail(slotNum: int = None) -> Optional[dict]:
    rp = method('getItemDetail', slotNum)
    if rp.peek() is None:
        return None
    return rp.take_dict()


def equipLeft():
    return method('equipLeft').check_bool_error()


def equipRight():
    return method('equipRight').check_bool_error()


def attack():
    return method('attack').check_bool_error()


def attackUp():
    return method('attackUp').check_bool_error()


def attackDown():
    return method('attackDown').check_bool_error()


def dig():
    return method('dig').check_bool_error()


def digUp():
    return method('digUp').check_bool_error()


def digDown():
    return method('digDown').check_bool_error()


def place(signText: str = None):
    return method('place', ser.nil_encode(signText)).check_bool_error()


def placeUp(signText: str = None):
    return method('placeUp', ser.nil_encode(signText)).check_bool_error()


def placeDown(signText: str = None):
    return method('placeDown', ser.nil_encode(signText)).check_bool_error()


def detect() -> bool:
    return method('detect').take_bool()


def detectUp() -> bool:
    return method('detectUp').take_bool()


def detectDown() -> bool:
    return method('detectDown').take_bool()


def inspect() -> Optional[dict]:
    return inspect_result(method('inspect'))


def inspectUp() -> Optional[dict]:
    return inspect_result(method('inspectUp'))


def inspectDown() -> Optional[dict]:
    return inspect_result(method('inspectDown'))


def compare() -> bool:
    return method('compare').take_bool()


def compareUp() -> bool:
    return method('compareUp').take_bool()


def compareDown() -> bool:
    return method('compareDown').take_bool()


def compareTo(slot: int) -> bool:
    return method('compareTo', slot).take_bool()


def drop(count: int = None):
    return method('drop', count).check_bool_error()


def dropUp(count: int = None):
    return method('dropUp', count).check_bool_error()


def dropDown(count: int = None):
    return method('dropDown', count).check_bool_error()


def suck(amount: int = None):
    return method('suck', amount).check_bool_error()


def suckUp(amount: int = None):
    return method('suckUp', amount).check_bool_error()


def suckDown(amount: int = None):
    return method('suckDown', amount).check_bool_error()


def refuel(quantity: int = None):
    return method('refuel', quantity).check_bool_error()


def getFuelLevel() -> int:
    return method('getFuelLevel').take_int()


def getFuelLimit() -> int:
    return method('getFuelLimit').take_int()


def transferTo(slot: int, quantity: int = None):
    return method('transferTo', slot, quantity).check_bool_error()



from uuid import UUID

from .errors import LuaException


def lua_table_to_list(x, length: int = None, low_index: int = 1):
    if not x:
        return [] if length is None else [None] * length
    assert all(map(lambda k: isinstance(k, int), x.keys()))
    # assert min(x.keys()) >= low_index
    dlen = max(x.keys()) - low_index + 1
    if length is not None:
        assert dlen <= length
    else:
        length = dlen
    return [x.get(i + low_index) for i in range(length)]


class ResultProc:
    def __init__(self, result, enc):
        self._v = result
        self._enc = enc
        self._i = 1

    def forward(self):
        self._i += 1

    def back(self):
        self._i -= 1

    def peek(self):
        return self._v.get(self._i)

    def peek_type(self):
        v = self.peek()
        if v is None:
            return None
        return type(v)

    def take(self):
        r = self.peek()
        self.forward()
        return r

    def take_none(self):
        x = self.take()
        assert x is None
        return x

    def take_bool(self):
        x = self.take()
        assert x is True or x is False
        return x

    def take_bool_coerce_nil(self):
        x = self.take()
        assert x is True or x is False or x is None
        return bool(x)

    def take_int(self):
        x = self.take()
        assert isinstance(x, int)
        assert not isinstance(x, bool)
        return x

    def take_number(self):
        x = self.take()
        assert isinstance(x, (int, float))
        assert not isinstance(x, bool)
        return x

    def take_bytes(self):
        x = self.take()
        assert isinstance(x, bytes)
        return x

    def take_string(self):
        return self.take_bytes().decode(self._enc)

    def take_unicode(self):
        # specific case in CC, reading unicode files
        return self.take_bytes().decode('utf-8')

    def take_uuid(self):
        return UUID(self.take_bytes().decode('ascii'))

    def take_dict(self, keys=None):
        x = self.take()
        assert isinstance(x, dict)
        if keys is None:
            return x
        return TableProc(x, keys)

    def take_list(self, length: int = None):
        return lua_table_to_list(self.take_dict(), length)

    def take_proc(self):
        x = self.take()
        assert isinstance(x, dict)
        return ResultProc(x, self._enc)

    def check_bool_error(self):
        success = self.take_bool()
        if not success:
            raise LuaException(self.take_string())

    def check_nil_error(self, allow_nil_nil=False):
        if self.peek() is None:
            self.forward()
            if allow_nil_nil and self.peek() is None:
                return
            raise LuaException(self.take_string())

    def take_option_int(self):
        if self.peek() is None:
            return None
        return self.take_int()

    def take_option_bytes(self):
        if self.peek() is None:
            return None
        return self.take_bytes()

    def take_option_string(self):
        if self.peek() is None:
            return None
        return self.take_string()

    def take_option_unicode(self):
        if self.peek() is None:
            return None
        return self.take_unicode()

    def take_option_string_bool(self):
        p = self.peek()
        if p is None or p is True or p is False:
            self.forward()
            return p
        return self.take_string()

    def take_list_of_strings(self, length: int = None):
        x = self.take_list(length)
        assert all(map(lambda v: isinstance(v, bytes), x))
        return [v.decode(self._enc) for v in x]

    def take_2d_int(self):
        x = self.take_list()
        x = [lua_table_to_list(item) for item in x]
        for row in x:
            for item in row:
                assert isinstance(item, int)
        return x


class TableProc(ResultProc):
    def __init__(self, result, keys):
        self._v = result
        self._keys = keys
        self._i = 0

    def peek(self):
        return self._v.get(self._keys[self._i])



from typing import Optional

from .. import ser
from ..lua import LuaNum
from ..sess import eval_lua_method_factory, get_current_greenlet


method = eval_lua_method_factory('os.')


__all__ = (
    'version',
    'getComputerID',
    'getComputerLabel',
    'setComputerLabel',
    'run',
    'captureEvent',
    'queueEvent',
    'clock',
    'time',
    'day',
    'epoch',
    'sleep',
    'startTimer',
    'cancelTimer',
    'setAlarm',
    'cancelAlarm',
    'shutdown',
    'reboot',
)


def version() -> str:
    return method('version').take_string()


def getComputerID() -> int:
    return method('getComputerID').take_int()


def getComputerLabel() -> Optional[str]:
    return method('getComputerLabel').take_option_string()


def setComputerLabel(label: Optional[str]):
    return method('setComputerLabel', ser.nil_encode(label)).take_none()


def run(environment: dict, programPath: str, *args: str):
    args = tuple(ser.encode(a) for a in args)
    return method('run', environment, ser.encode(programPath), *args).take_bool()


def captureEvent(event: str):
    event = ser.encode(event)
    glet = get_current_greenlet().cc_greenlet
    sess = glet._sess
    evr = sess._evr
    evr.sub(glet._task_id, event)
    try:
        while True:
            val = evr.get_from_stack(glet._task_id, event)
            if val is None:
                res = sess._server_greenlet.switch()
                assert res == 'event'
            else:
                yield val
    finally:
        evr.unsub(glet._task_id, event)


def queueEvent(event: str, *params):
    return method('queueEvent', ser.encode(event), *params).take_none()


def clock() -> LuaNum:
    # number of game ticks * 0.05, roughly seconds
    return method('clock').take_number()


# regarding ingame parameter below:
# python has great stdlib to deal with real current time
# we keep here only in-game time methods and parameters

def time() -> LuaNum:
    # in hours 0..24
    return method('time', b'ingame').take_number()


def day() -> int:
    return method('day', b'ingame').take_int()


def epoch() -> int:
    return method('epoch', b'ingame').take_int()


def sleep(seconds: LuaNum):
    return method('sleep', seconds).take_none()


def startTimer(timeout: LuaNum) -> int:
    return method('startTimer', timeout).take_int()


def cancelTimer(timerID: int):
    return method('cancelTimer', timerID).take_none()


def setAlarm(time: LuaNum) -> int:
    # takes time of the day in hours 0..24
    # returns integer alarmID
    return method('setAlarm', time).take_int()


def cancelAlarm(alarmID: int):
    return method('cancelAlarm', alarmID).take_none()


def shutdown():
    return method('shutdown').take_none()


def reboot():
    return method('reboot').take_none()



from typing import Any, List

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('settings.')


__all__ = (
    'define',
    'undefine',
    'getDetails',
    'set',
    'get',
    'unset',
    'clear',
    'getNames',
    'load',
    'save',
)


def define(name: str, description: str = None, default: Any = None, type: str = None):
    options = {}
    if description is not None:
        options[b'description'] = ser.encode(description)
    if default is not None:
        options[b'default'] = default
    if type is not None:
        options[b'type'] = ser.encode(type)
    return method('define', ser.encode(name), options).take_none()


def undefine(name: str):
    return method('undefine', ser.encode(name)).take_none()


def getDetails(name: str) -> dict:
    tp = method('getDetails', ser.encode(name)).take_dict((
        b'changed',
        b'description',
        b'default',
        b'type',
        b'value',
    ))
    r = {}
    r['changed'] = tp.take_bool()
    for k, v in [
        ('description', tp.take_option_string()),
        ('default', tp.take()),
        ('type', tp.take_option_string()),
        ('value', tp.take()),
    ]:
        if v is not None:
            r[k] = v
    return r


def set(name: str, value: Any):
    return method('set', ser.encode(name), value).take_none()


def get(name: str, default: Any = None) -> Any:
    return method('get', ser.encode(name), default).take()


def unset(name: str):
    return method('unset', ser.encode(name)).take_none()


def clear():
    return method('clear').take_none()


def getNames() -> List[str]:
    return method('getNames').take_list_of_strings()


def load(path: str = None) -> bool:
    return method('load', ser.nil_encode(path)).take_bool()


def save(path: str = None) -> bool:
    return method('save', ser.nil_encode(path)).take_bool()



from typing import List

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('paintutils.')


__all__ = (
    'parseImage',
    'loadImage',
    'drawPixel',
    'drawLine',
    'drawBox',
    'drawFilledBox',
    'drawImage',
)


def parseImage(data: bytes) -> List[List[int]]:
    return method('parseImage', data).take_2d_int()


def loadImage(path: str) -> List[List[int]]:
    return method('loadImage', ser.encode(path)).take_2d_int()


def drawPixel(x: int, y: int, color: int = None):
    return method('drawPixel', x, y, color).take_none()


def drawLine(startX: int, startY: int, endX: int, endY: int, color: int = None):
    return method('drawLine', startX, startY, endX, endY, color).take_none()


def drawBox(startX: int, startY: int, endX: int, endY: int, color: int = None):
    return method('drawBox', startX, startY, endX, endY, color).take_none()


def drawFilledBox(startX: int, startY: int, endX: int, endY: int, color: int = None):
    return method('drawFilledBox', startX, startY, endX, endY, color).take_none()


def drawImage(image: List[List[int]], xPos: int, yPos: int):
    return method('drawImage', image, xPos, yPos).take_none()



from typing import List

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('redstone.')


__all__ = (
    'getSides',
    'getInput',
    'setOutput',
    'getOutput',
    'getAnalogInput',
    'setAnalogOutput',
    'getAnalogOutput',
    'getBundledInput',
    'setBundledOutput',
    'getBundledOutput',
    'testBundledInput',
)


def getSides() -> List[str]:
    return method('getSides').take_list_of_strings()


def getInput(side: str) -> bool:
    return method('getInput', ser.encode(side)).take_bool()


def setOutput(side: str, value: bool):
    return method('setOutput', ser.encode(side), value).take_none()


def getOutput(side: str) -> bool:
    return method('getOutput', ser.encode(side)).take_bool()


def getAnalogInput(side: str) -> int:
    return method('getAnalogInput', ser.encode(side)).take_int()


def setAnalogOutput(side: str, strength: int):
    return method('setAnalogOutput', ser.encode(side), strength).take_none()


def getAnalogOutput(side: str) -> int:
    return method('getAnalogOutput', ser.encode(side)).take_int()


# bundled cables are not available in vanilla

def getBundledInput(side: str) -> int:
    return method('getBundledInput', ser.encode(side)).take_int()


def setBundledOutput(side: str, colors: int):
    return method('setBundledOutput', ser.encode(side), colors).take_none()


def getBundledOutput(side: str) -> int:
    return method('getBundledOutput', ser.encode(side)).take_int()


def testBundledInput(side: str, color: int) -> bool:
    return method('testBundledInput', ser.encode(side), color).take_bool()



from typing import Optional, List

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('help.')


__all__ = (
    'path',
    'setPath',
    'lookup',
    'topics',
    'completeTopic',
)


def path() -> str:
    return method('path').take_string()


def setPath(path: str):
    return method('setPath', ser.encode(path)).take_none()


def lookup(topic: str) -> Optional[str]:
    return method('lookup', ser.encode(topic)).take_option_string()


def topics() -> List[str]:
    return method('topics').take_list_of_strings()


def completeTopic(topicPrefix: str) -> List[str]:
    return method('completeTopic', ser.encode(topicPrefix)).take_list_of_strings()



from typing import Any, List, Optional, Tuple, Union

from .. import ser
from ..lua import LuaNum
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('rednet.')


__all__ = (
    'CHANNEL_REPEAT',
    'CHANNEL_BROADCAST',
    'open',
    'close',
    'send',
    'broadcast',
    'receive',
    'isOpen',
    'host',
    'unhost',
    'lookup',
)


CHANNEL_REPEAT = 65533
CHANNEL_BROADCAST = 65535


def open(side: str):
    return method('open', ser.encode(side)).take_none()


def close(side: str = None):
    return method('close', ser.nil_encode(side)).take_none()


def send(receiverID: int, message: Any, protocol: str = None) -> bool:
    return method('send', receiverID, message, ser.nil_encode(protocol)).take_bool()


def broadcast(message: Any, protocol: str = None):
    return method('broadcast', message, ser.nil_encode(protocol)).take_none()


def receive(
    protocolFilter: str = None, timeout: LuaNum = None,
) -> Optional[Tuple[int, Any, Optional[str]]]:
    rp = method('receive', ser.nil_encode(protocolFilter), timeout)
    if rp.peek() is None:
        return None
    return (rp.take_int(), rp.take(), rp.take_option_string())


def isOpen(side: str = None) -> bool:
    return method('isOpen', ser.nil_encode(side)).take_bool()


def host(protocol: str, hostname: str):
    return method('host', ser.encode(protocol), ser.encode(hostname)).take_none()


def unhost(protocol: str):
    return method('unhost', ser.encode(protocol)).take_none()


def lookup(protocol: str, hostname: str = None) -> Union[Optional[int], List[int]]:
    rp = method('lookup', ser.encode(protocol), ser.nil_encode(hostname))
    if hostname is None:
        r = []
        while rp.peek() is not None:
            r.append(rp.take_int())
        return r
    else:
        return rp.take_option_int()



from typing import Tuple

from .. import ser
from ..lua import LuaExpr


class TermMixin:
    def write(self, text: str):
        return self._method('write', ser.dirty_encode(str(text),'utf-8')).take_none()

    def blit(self, text: str, textColors: bytes, backgroundColors: bytes):
        return self._method('blit', ser.dirty_encode(text), textColors, backgroundColors).take_none()

    def clear(self):
        return self._method('clear').take_none()

    def clearLine(self):
        return self._method('clearLine').take_none()

    def getCursorPos(self) -> Tuple[int, int]:
        rp = self._method('getCursorPos')
        return tuple(rp.take_int() for _ in range(2))

    def setCursorPos(self, x: int, y: int):
        return self._method('setCursorPos', x, y).take_none()

    def getCursorBlink(self) -> bool:
        return self._method('getCursorBlink').take_bool()

    def setCursorBlink(self, value: bool):
        return self._method('setCursorBlink', value).take_none()

    def isColor(self) -> bool:
        return self._method('isColor').take_bool()

    def getSize(self) -> Tuple[int, int]:
        rp = self._method('getSize')
        return tuple(rp.take_int() for _ in range(2))

    def scroll(self, lines: int):
        return self._method('scroll', lines).take_none()

    def setTextColor(self, colorID: int):
        return self._method('setTextColor', colorID).take_none()

    def getTextColor(self) -> int:
        return self._method('getTextColor').take_int()

    def setBackgroundColor(self, colorID: int):
        return self._method('setBackgroundColor', colorID).take_none()

    def getBackgroundColor(self) -> int:
        return self._method('getBackgroundColor').take_int()

    def getPaletteColor(self, colorID: int) -> Tuple[float, float, float]:
        rp = self._method('getPaletteColor', colorID)
        return tuple(rp.take_number() for _ in range(3))

    def setPaletteColor(self, colorID: int, r: float, g: float, b: float):
        return self._method('setPaletteColor', colorID, r, g, b).take_none()


class TermTarget(LuaExpr):
    def __init__(self, code):
        self._code = code

    def get_expr_code(self):
        return self._code



import argparse
import asyncio
from os.path import join, dirname, abspath

from aiohttp import web, WSMsgType

from . import ser, sess
from .rproc import lua_table_to_list


THIS_DIR = dirname(abspath(__file__))
LUA_FILE = join(THIS_DIR, 'back.lua')
PROTO_VERSION = 5
PROTO_ERROR = b'C' + ser.serialize(b'protocol error', 'ascii')


async def _bin_messages(ws):
    async for msg in ws:
        if msg.type != WSMsgType.BINARY:
            continue
        yield msg


def protocol(send, sess_cls=sess.CCSession, oc=False):
    # handle first frame
    msg = yield
    msg = ser.dcmditer(msg)

    action = next(msg)
    if action != b'0':
        send(PROTO_ERROR)
        return

    version = next(msg)
    if version != PROTO_VERSION:
        send(b'C' + ser.serialize(
            'protocol version mismatch (expected {}, got {}), redownload py'.format(
                PROTO_VERSION, version,
            ), 'ascii'))
        return
    args = lua_table_to_list(next(msg))
    path, code = None, None
    try:
        path = next(msg)
        code = next(msg)
    except StopIteration:
        pass
    sess = sess_cls(send, oc=oc)
    if code is not None:
        sess.run_program(args, path, code)
    else:
        sess.run_repl()

    # handle the rest of frames
    while True:
        msg = yield
        msg = ser.dcmditer(msg)
        action = next(msg)
        if action == b'E':
            sess.on_event(next(msg), lua_table_to_list(next(msg)))
        elif action == b'T':
            sess.on_task_result(next(msg), next(msg))
        else:
            send(PROTO_ERROR)
            return


class CCApplication(web.Application):
    async def ws(self, request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        squeue = []
        pgen = self['protocol_factory'](squeue.append, oc=False)
        next(pgen)
        mustquit = False
        async for msg in _bin_messages(ws):
            if msg.type != WSMsgType.BINARY:
                continue

            try:
                pgen.send(msg.data)
            except StopIteration:
                mustquit = True

            for m in squeue:
                await ws.send_bytes(m)
            squeue.clear()

            if mustquit:
                break
        return ws

    async def tcp(self, reader, writer):
        squeue = []

        def send(m):
            squeue.append(m)

        pgen = self['protocol_factory'](squeue.append, oc=True)
        next(pgen)
        mustquit = False
        while True:
            try:
                frame_size = int.from_bytes(
                    await reader.readexactly(3), 'big')
                frame = await reader.readexactly(frame_size)
            except asyncio.IncompleteReadError:
                break

            try:
                pgen.send(frame)
            except StopIteration:
                mustquit = True

            for m in squeue:
                writer.write(len(m).to_bytes(3, 'big'))
                writer.write(m)
            squeue.clear()
            await writer.drain()

            if mustquit:
                break

        writer.close()
        await writer.wait_closed()

    @staticmethod
    def backdoor(request):
        with open(LUA_FILE, 'r') as f:
            fcont = f.read()
        webhost = tcphost = request.host
        if ':' in request.host:
            tcphost = tcphost.rsplit(':', 1)[0]
        else:
            # fix for malformed Host header
            webhost += ':{}'.format(request.app['port'])
        return web.Response(text=(
            fcont
            .replace('__cc_url__', 'ws://{}/ws/'.format(webhost))
            .replace('__oc_host__', tcphost)
            .replace('__oc_port__', str(request.app['oc_port']))
        ))

    def setup_routes(self):
        self.router.add_get('/', self.backdoor)
        self.router.add_get('/ws/', self.ws)


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', default='0.0.0.0')
    parser.add_argument(
        '--port', type=int, default=8000,
        help='Web (for wget) & websocket (for computercraft) port')
    parser.add_argument(
        '--oc-port', type=int, default=8001,
        help='Raw TCP port for opencomputers')
    parser.add_argument(
        '--capture', type=str, default=None,
        help='Capture test data into a file')
    return parser


def main():
    args = create_parser().parse_args()
    app = CCApplication()
    app['port'] = args.port
    app['oc_port'] = args.oc_port
    app['protocol_factory'] = protocol
    app.setup_routes()

    async def tcp_server(app):
        server = await asyncio.start_server(app.tcp, args.host, args.oc_port)
        async with server:
            yield

    async def capture(app):
        with open(args.capture, 'wb') as f:
            def protocol_factory(send, sess_cls=sess.CCSession, oc=False):
                def write_frame(t, m):
                    ln = str(len(m)).encode('ascii')
                    f.write(t + ln + b':' + m + b'\n')

                def send_wrap(m):
                    write_frame(b'S', m)
                    return send(m)

                p = protocol(send_wrap, sess_cls=sess_cls, oc=oc)

                def pgen():
                    next(p)
                    while True:
                        m = yield
                        write_frame(b'R', m)
                        p.send(m)

                return pgen()

            app['protocol_factory'] = protocol_factory
            yield

    app.cleanup_ctx.append(tcp_server)
    if args.capture is not None:
        sess.python_version = lambda: '<VERSION>'
        app.cleanup_ctx.append(capture)

    with sess.patch_std_files():
        web.run_app(app, host=args.host, port=args.port)


if __name__ == '__main__':
    main()


# TODO: move greenlets into separate thread
# to prevent hanging

"""
import ctypes
import time
import threading

def thmain():
    # hangs
    while True:
        pass

th = threading.Thread(target=thmain)
th.start()
time.sleep(5)
ctypes.pythonapi.PyThreadState_SetAsyncExc(
    ctypes.c_long(th.ident),
    ctypes.py_object(TimeoutError))
time.sleep(3)
"""



from typing import List, Dict, Optional

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('shell.')


__all__ = (
    'exit',
    'dir',
    'setDir',
    'path',
    'setPath',
    'resolve',
    'resolveProgram',
    'aliases',
    'setAlias',
    'clearAlias',
    'programs',
    'getRunningProgram',
    'run',
    'execute',
    'openTab',
    'switchTab',
    'complete',
    'completeProgram',
)


def exit():
    return method('exit').take_none()


def dir() -> str:
    return method('dir').take_string()


def setDir(path: str):
    return method('setDir', ser.encode(path)).take_none()


def path() -> str:
    return method('path').take_string()


def setPath(path: str):
    return method('setPath', ser.encode(path)).take_none()


def resolve(localPath: str) -> str:
    return method('resolve', ser.encode(localPath)).take_string()


def resolveProgram(name: str) -> Optional[str]:
    return method('resolveProgram', ser.encode(name)).take_option_string()


def aliases() -> Dict[str, str]:
    d = method('aliases').take_dict()
    return {ser.decode(k): ser.decode(v) for k, v in d.items()}


def setAlias(alias: str, program: str):
    return method('setAlias', ser.encode(alias), ser.encode(program)).take_none()


def clearAlias(alias: str):
    return method('clearAlias', ser.encode(alias)).take_none()


def programs(showHidden: bool = None) -> List[str]:
    return method('programs', showHidden).take_list_of_strings()


def getRunningProgram() -> str:
    return method('getRunningProgram').take_string()


def run(command: str, *args: str) -> bool:
    args = tuple(ser.encode(a) for a in args)
    return method('run', ser.encode(command), *args).take_bool()


def execute(command: str, *args: str) -> bool:
    args = tuple(ser.encode(a) for a in args)
    return method('execute', ser.encode(command), *args).take_bool()


def openTab(command: str, *args: str) -> int:
    args = tuple(ser.encode(a) for a in args)
    return method('openTab', ser.encode(command), *args).take_int()


def switchTab(tabID: int):
    return method('switchTab', tabID).take_none()


def complete(prefix: str) -> List[str]:
    return method('complete', ser.encode(prefix)).take_list_of_strings()


def completeProgram(prefix: str) -> List[str]:
    return method('completeProgram', ser.encode(prefix)).take_list_of_strings()

# TODO: ?
# these functions won't be implemented
# it's far better to keep this in lua code

# setCompletionFunction
# getCompletionInfo

# we can create callbacks to python code, but this will require
# connection to python, and will break the shell if python disconnects



from typing import Any, Tuple, Union
from uuid import UUID

from . import lua


__all__ = (
    'serialize',
    'deserialize',
)


_ENC = 'latin1'
# encoding fast check
assert [bytes([i]) for i in range(256)] == [chr(i).encode(_ENC) for i in range(256)]


def encode(s: str) -> bytes:
    return s.encode(_ENC)


def u_encode(s: Union[str, UUID, None]) -> Union[bytes, UUID, None]:
    if isinstance(s, str):
        return s.encode('utf-8')
    return s


def nil_encode(s):
    if s is None:
        return None
    return encode(s)


def dirty_encode(s: str, enc: str) -> bytes:
    return s.encode(enc, errors='replace')


def decode(b):
    return b.decode(_ENC)


def u_decode(b):
    return b.decode('utf-8')


def serialize(v: Any, encoding: str) -> bytes:
    if v is None:
        return b'N'
    elif v is False:
        return b'F'
    elif v is True:
        return b'T'
    elif isinstance(v, (int, float)):
        return '[{}]'.format(v).encode('ascii')
    elif isinstance(v, bytes):
        return '<{}>'.format(len(v)).encode('ascii') + v
    elif isinstance(v, UUID):
        return serialize(str(v).encode('ascii'), encoding)
    elif isinstance(v, str):
        return serialize(v.encode(encoding), encoding)
    elif isinstance(v, (list, tuple)):
        items = []
        for k, x in enumerate(v, start=1):
            items.append(b':' + serialize(k, encoding) + serialize(x, encoding))
        return b'{' + b''.join(items) + b'}'
    elif isinstance(v, dict):
        items = []
        for k, x in v.items():
            items.append(b':' + serialize(k, encoding) + serialize(x, encoding))
        return b'{' + b''.join(items) + b'}'
    elif isinstance(v, lua.LuaExpr):
        code = v.get_expr_code()
        return 'E{}>'.format(len(code)).encode('ascii') + code
    else:
        raise ValueError('Value can\'t be serialized: {}'.format(repr(v)))


def _deserialize(b: bytes, _idx: int) -> Tuple[Any, int]:
    tok = b[_idx]
    _idx += 1
    if tok == 78:  # N
        return None, _idx
    elif tok == 70:  # F
        return False, _idx
    elif tok == 84:  # T
        return True, _idx
    elif tok == 91:  # [
        newidx = b.index(b']', _idx)
        f = float(b[_idx:newidx])
        if f.is_integer():
            f = int(f)
        return f, newidx + 1
    elif tok == 60:  # <
        newidx = b.index(b'>', _idx)
        ln = int(b[_idx:newidx])
        return b[newidx + 1:newidx + 1 + ln], newidx + 1 + ln
    elif tok == 123:  # {
        r = {}
        while True:
            tok = b[_idx]
            _idx += 1
            if tok == 125:  # }
                break
            key, _idx = _deserialize(b, _idx)
            value, _idx = _deserialize(b, _idx)
            r[key] = value
        return r, _idx
    else:
        raise ValueError


def deserialize(b: bytes) -> Any:
    return _deserialize(b, 0)[0]


def dcmditer(b: bytes):
    yield b[0:1]
    idx = 1
    while idx < len(b):
        chunk, idx = _deserialize(b, idx)
        yield chunk



import asyncio
import string
import sys
from code import InteractiveConsole
from collections import deque
from contextlib import contextmanager
from functools import partial
from importlib import import_module
from importlib.abc import MetaPathFinder, Loader
from importlib.machinery import ModuleSpec
from itertools import count
from platform import python_version
from traceback import format_exc
from types import ModuleType

from greenlet import greenlet, getcurrent as get_current_greenlet

from . import rproc, ser


__all__ = (
    'CCSession',
    'get_current_session',
    'eval_lua',
    'lua_context_object',
)


def debug(*args):
    sys.__stdout__.write(' '.join(map(str, args)) + '\n')
    sys.__stdout__.flush()


DIGITS = (string.digits + string.ascii_lowercase).encode('ascii')


def base36(n: int) -> bytes:
    r = bytearray()
    while n:
        r.append(DIGITS[n % 36])
        n //= 36
    return bytes(r)


def _is_global_greenlet():
    return not hasattr(get_current_greenlet(), 'cc_greenlet')


def get_current_session():
    try:
        return get_current_greenlet().cc_greenlet._sess
    except AttributeError:
        raise RuntimeError('Computercraft function was called outside context')


class StdFileProxy:
    def __init__(self, native, err):
        self._native = native
        self._err = err

    def read(self, size=-1):
        if _is_global_greenlet():
            return self._native.read(size)
        else:
            raise RuntimeError(
                "Computercraft environment doesn't support stdin read method")

    def readline(self, size=-1):
        if _is_global_greenlet():
            return self._native.readline(size)
        else:
            if size is not None and size >= 0:
                raise RuntimeError(
                    "Computercraft environment doesn't support "
                    "stdin readline method with parameter")
            r = eval_lua(b'G:io:M:read')
            if r.peek() is None:
                return ''  # press ctrl+D in OC
            if r.peek() is False:
                r.take()   # press ctrl+C in OC
                eval_lua(b'io.stderr:write(...)', r.take_bytes(), immediate=True)
                return ''
            return r.take_string() + '\n'

    def write(self, s):
        if _is_global_greenlet():
            return self._native.write(s)
        else:
            s = ser.dirty_encode(s, get_current_session()._enc)
            if self._err:
                return eval_lua(b'io.stderr:write(...)', s).take_none()
            else:
                return eval_lua(b'io.write(...)', s).take_none()

    def fileno(self):
        if _is_global_greenlet():
            return self._native.fileno()
        else:
            # preventing use of gnu readline here
            # https://github.com/python/cpython/blob/master/Python/bltinmodule.c#L1970
            raise AttributeError

    def __getattr__(self, name):
        return getattr(self._native, name)


class ComputerCraftFinder(MetaPathFinder):
    @staticmethod
    def find_spec(fullname, path, target=None):
        if fullname == 'cc':
            return ModuleSpec(fullname, ComputerCraftLoader, is_package=True)
        if fullname.startswith('cc.'):
            return ModuleSpec(fullname, ComputerCraftLoader, is_package=False)
        if fullname == 'oc':
            return ModuleSpec(fullname, OpenComputersLoader, is_package=True)
        if fullname.startswith('oc.'):
            return ModuleSpec(fullname, OpenComputersLoader, is_package=False)


class ComputerCraftLoader(Loader):
    @staticmethod
    def create_module(spec):
        sn = spec.name.split('.', 1)
        assert sn[0] == 'cc'
        if len(sn) == 1:
            sn.append('_pkg')
        rawmod = import_module('.' + sn[1], 'computercraft.subapis')
        mod = ModuleType(spec.name)
        for k in rawmod.__all__:
            setattr(mod, k, getattr(rawmod, k))
        return mod

    @staticmethod
    def exec_module(module):
        pass


class OpenComputersLoader(Loader):
    @staticmethod
    def create_module(spec):
        sn = spec.name.split('.', 1)
        assert sn[0] == 'oc'
        if len(sn) == 1:
            sn.append('_pkg')
        rawmod = import_module('.' + sn[1], 'computercraft.oc')
        mod = getattr(rawmod, '__replace_module__', None)
        if mod is not None:
            return mod
        mod = ModuleType(spec.name)
        for k in rawmod.__all__:
            setattr(mod, k, getattr(rawmod, k))
        return mod

    @staticmethod
    def exec_module(module):
        pass


def install_import_hook():
    import sys
    sys.meta_path.append(ComputerCraftFinder)


install_import_hook()


@contextmanager
def patch_std_files():
    pin, pout, perr = sys.stdin, sys.stdout, sys.stderr
    sys.stdin = StdFileProxy(pin, False)
    sys.stdout = StdFileProxy(pout, False)
    sys.stderr = StdFileProxy(perr, True)
    try:
        yield
    finally:
        sys.stdin, sys.stdout, sys.stderr = pin, pout, perr


def eval_lua(lua_code, *params, immediate=False):
    sess = get_current_session()
    # assert isinstance(lua_code, bytes)
    request = (
        (b'I' if immediate else b'T')
        + ser.serialize(lua_code, sess._enc)
        + ser.serialize(params, sess._enc)
    )
    result = sess._server_greenlet.switch(request)
    rp = rproc.ResultProc(ser.deserialize(result), sess._enc)
    if not immediate:
        rp.check_bool_error()
    return rp


@contextmanager
def lua_context_object(
    create_expr: bytes,
    create_params: tuple,
    finalizer_template: bytes = b'',
    prefix: bytes = b'',
):
    sess = get_current_session()
    fid = sess.create_task_id()
    r = eval_lua(
        prefix
        + b'return(function(n,...)local o,e='
        + create_expr
        + b';if o then temp[n]=o;return true;end'
        + b';return o,e;end)(...)', fid, *create_params)
    r.check_nil_error()
    assert r.take_bool() is True
    try:
        yield fid
    finally:
        eval_lua(
            b'local n=...;'
            + finalizer_template.replace(b'{e}', b'temp[n]')
            + b';temp[n]=nil', fid)


class ContextObject:
    # TODO: support serializing this object

    def __init__(self, fid):
        self._fid = fid

    def _call(self, method: bytes, *args):
        return eval_lua(
            b'return(function(n,...)return temp[n]'
            + method
            + b'(...);end)(...)', self._fid, *args)


def eval_lua_method_factory(obj):
    # TODO: remove this
    # eval strings must be cacheable
    def method(name, *params):
        code = 'return ' + obj + name + '(...)'
        return eval_lua(code, *params)
    return method


class CCGreenlet:
    def __init__(self, body_fn, sess=None):
        if sess is None:
            self._sess = get_current_session()
        else:
            self._sess = sess

        self._task_id = self._sess.create_task_id()
        self._sess._greenlets[self._task_id] = self

        parent_g = get_current_greenlet()
        if parent_g is self._sess._server_greenlet:
            self._parent = None
        else:
            self._parent = parent_g.cc_greenlet
            self._parent._children.add(self._task_id)

        self._children = set()
        self._g = greenlet(body_fn)
        self._g.cc_greenlet = self

    def detach_children(self):
        if self._children:
            ch = list(self._children)
            self._children.clear()
            self._sess.drop(ch)

    def _on_death(self, error=None):
        self._sess._greenlets.pop(self._task_id, None)
        self.detach_children()
        if error is not None:
            if error is True:
                error = None
            else:
                error = ser.dirty_encode(error, self._sess._enc)
            self._sess._sender(b'C' + ser.serialize(error, self._sess._enc))
        if self._parent is not None:
            self._parent._children.discard(self._task_id)

    def defer_switch(self, *args, **kwargs):
        asyncio.get_running_loop().call_soon(
            partial(self.switch, *args, **kwargs))

    def switch(self, *args, **kwargs):
        # switch must be called from server greenlet
        assert get_current_greenlet() is self._sess._server_greenlet
        try:
            task = self._g.switch(*args, **kwargs)
        except SystemExit:
            self._on_death(True)
            return
        except Exception:
            self._on_death(format_exc(limit=None, chain=False))
            return

        # lua_eval call or simply idle
        if isinstance(task, bytes):
            x = self
            while x._g.dead:
                x = x._parent
            self._sess._sender(
                task[0:1] + ser.serialize(x._task_id, 'ascii') + task[1:])

        if self._g.dead:
            if self._parent is None:
                self._on_death(True)
            else:
                self._on_death()


class CCEventRouter:
    def __init__(self, on_first_sub, on_last_unsub, resume_task):
        self._stacks = {}
        self._active = {}
        self._on_first_sub = on_first_sub
        self._on_last_unsub = on_last_unsub
        self._resume_task = resume_task

    def sub(self, task_id, event):
        if event not in self._stacks:
            self._stacks[event] = {}
            self._on_first_sub(event)
        se = self._stacks[event]
        if task_id in se:
            raise Exception('Same task subscribes to the same event twice')
        se[task_id] = deque()

    def unsub(self, task_id, event):
        if event not in self._stacks:
            return
        self._stacks[event].pop(task_id, None)
        if len(self._stacks[event]) == 0:
            self._on_last_unsub(event)
            del self._stacks[event]

    def on_event(self, event, params):
        if event not in self._stacks:
            self._on_last_unsub(event)
            return
        for task_id, queue in self._stacks[event].items():
            queue.append(params)
            if self._active.get(task_id) == event:
                self._set_task_status(task_id, event, False)
                self._resume_task(task_id)

    def get_from_stack(self, task_id, event):
        queue = self._stacks[event][task_id]
        try:
            return queue.popleft()
        except IndexError:
            self._set_task_status(task_id, event, True)
            return None

    def _set_task_status(self, task_id, event, waits: bool):
        if waits:
            self._active[task_id] = event
        else:
            self._active.pop(task_id, None)


class CCSession:
    def __init__(self, sender, oc=False):
        self._tid_allocator = map(base36, count(start=1))
        self._sender = sender
        self._oc = oc
        self._enc = 'utf-8' if oc else 'latin1'
        self._greenlets = {}
        self._server_greenlet = get_current_greenlet()
        self._program_greenlet = None
        self._evr = CCEventRouter(
            lambda event: self._sender(b'S' + ser.serialize(event, self._enc)),
            lambda event: self._sender(b'U' + ser.serialize(event, self._enc)),
            lambda task_id: self._greenlets[task_id].defer_switch('event'),
        )

    def on_task_result(self, task_id, result):
        assert get_current_greenlet() is self._server_greenlet
        if task_id not in self._greenlets:
            # ignore for dropped tasks
            return
        self._greenlets[task_id].switch(result)

    def on_event(self, event, params):
        self._evr.on_event(event, params)

    def create_task_id(self):
        return next(self._tid_allocator)

    def drop(self, task_ids):
        def collect(task_id):
            yield task_id
            g = self._greenlets.pop(task_id)
            for tid in g._children:
                yield from collect(tid)

        all_tids = []
        for task_id in task_ids:
            all_tids.extend(collect(task_id))

        self._sender(b'D' + b''.join(
            ser.serialize(tid, self._enc) for tid in all_tids))

    def _run_sandboxed_greenlet(self, fn):
        self._program_greenlet = CCGreenlet(fn, sess=self)
        self._program_greenlet.switch()

    def run_program(self, args, path, code):
        def _run_program():
            cc = compile(code, path, 'exec')
            exec(cc, {'__file__': path, 'args': args})

        self._run_sandboxed_greenlet(_run_program)

    def run_repl(self):
        def _repl():
            InteractiveConsole(locals={}).interact(
                banner='Python {}'.format(python_version()),
                exitmsg='',
            )

        self._run_sandboxed_greenlet(_repl)



from contextlib import contextmanager
from typing import Tuple

from .base import BaseSubAPI
from .mixins import TermMixin, TermTarget
from ..sess import eval_lua_method_factory, lua_context_object


class TermAPI(BaseSubAPI, TermMixin):
    pass


method = eval_lua_method_factory('term.')
tapi = TermAPI('term')


__all__ = (
    'write',
    'blit',
    'clear',
    'clearLine',
    'getCursorPos',
    'setCursorPos',
    'getCursorBlink',
    'setCursorBlink',
    'isColor',
    'getSize',
    'scroll',
    'setTextColor',
    'getTextColor',
    'setBackgroundColor',
    'getBackgroundColor',
    'getPaletteColor',
    'setPaletteColor',
    'nativePaletteColor',
    'redirect',
    'get_current_target',
    'get_native_target',
)


write = tapi.write
blit = tapi.blit
clear = tapi.clear
clearLine = tapi.clearLine
getCursorPos = tapi.getCursorPos
setCursorPos = tapi.setCursorPos
getCursorBlink = tapi.getCursorBlink
setCursorBlink = tapi.setCursorBlink
isColor = tapi.isColor
getSize = tapi.getSize
scroll = tapi.scroll
setTextColor = tapi.setTextColor
getTextColor = tapi.getTextColor
setBackgroundColor = tapi.setBackgroundColor
getBackgroundColor = tapi.getBackgroundColor
getPaletteColor = tapi.getPaletteColor
setPaletteColor = tapi.setPaletteColor


def nativePaletteColor(colorID: int) -> Tuple[float, float, float]:
    rp = method('nativePaletteColor', colorID)
    return tuple(rp.take_number() for _ in range(3))


@contextmanager
def redirect(target: TermTarget):
    with lua_context_object(
        'term.redirect(...)',
        (target, ),
        'term.redirect({e})',
    ):
        yield


def get_current_target() -> TermTarget:
    return TermTarget('term.current()')


def get_native_target() -> TermTarget:
    return TermTarget('term.native()')



from typing import Tuple

from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('colors.')


__all__ = (
    'white',
    'orange',
    'magenta',
    'lightBlue',
    'yellow',
    'lime',
    'pink',
    'gray',
    'lightGray',
    'cyan',
    'purple',
    'blue',
    'brown',
    'green',
    'red',
    'black',
    'combine',
    'subtract',
    'test',
    'packRGB',
    'unpackRGB',
    'chars',
    'iter_colors',
)


white = 0x1
orange = 0x2
magenta = 0x4
lightBlue = 0x8
yellow = 0x10
lime = 0x20
pink = 0x40
gray = 0x80
lightGray = 0x100
cyan = 0x200
purple = 0x400
blue = 0x800
brown = 0x1000
green = 0x2000
red = 0x4000
black = 0x8000


# combine, subtract and test are mostly for redstone.setBundledOutput

def combine(*colors: int) -> int:
    return method('combine', *colors).take_int()


def subtract(color_set: int, *colors: int) -> int:
    return method('subtract', color_set, *colors).take_int()


def test(colors: int, color: int) -> bool:
    return method('test', colors, color).take_bool()


def packRGB(r: float, g: float, b: float) -> int:
    return method('packRGB', r, g, b).take_int()


def unpackRGB(rgb: int) -> Tuple[float, float, float]:
    rp = method('unpackRGB', rgb)
    return tuple(rp.take_number() for _ in range(3))


# use these chars for term.blit
chars = {
    '0': white,
    '1': orange,
    '2': magenta,
    '3': lightBlue,
    '4': yellow,
    '5': lime,
    '6': pink,
    '7': gray,
    '8': lightGray,
    '9': cyan,
    'a': purple,
    'b': blue,
    'c': brown,
    'd': green,
    'e': red,
    'f': black,
}


def iter_colors():
    for c in chars.values():
        yield c



from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('pocket.')


__all__ = (
    'equipBack',
    'unequipBack',
)


def equipBack():
    return method('equipBack').check_bool_error()


def unequipBack():
    return method('unequipBack').check_bool_error()



from ..lua import LuaExpr
from ..sess import eval_lua


class BaseSubAPI(LuaExpr):
    def __init__(self, lua_expr):
        self._lua_expr = lua_expr

    def get_expr_code(self):
        return self._lua_expr

    def _method(self, name, *params):
        code = 'return {}.{}(...)'.format(self.get_expr_code(), name)
        return eval_lua(code, *params)
